@@ -22,15 +22,15 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with ldap3 in the COPYING and COPYING.LESSER files.
 # If not, see <http://www.gnu.org/licenses/>.
-
+from copy import deepcopy
 from os import linesep
-from threading import RLock
+from threading import RLock, Lock
 from functools import reduce
 import json
 
 from .. import ANONYMOUS, SIMPLE, SASL, MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE, get_config_parameter, DEREF_ALWAYS, \
     SUBTREE, ASYNC, SYNC, NO_ATTRIBUTES, ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES, MODIFY_INCREMENT, LDIF, ASYNC_STREAM, \
-    RESTARTABLE, ROUND_ROBIN, REUSABLE, AUTO_BIND_NONE, AUTO_BIND_TLS_BEFORE_BIND, AUTO_BIND_TLS_AFTER_BIND, AUTO_BIND_NO_TLS, \
+    RESTARTABLE, ROUND_ROBIN, REUSABLE, AUTO_BIND_DEFAULT, AUTO_BIND_NONE, AUTO_BIND_TLS_BEFORE_BIND, AUTO_BIND_TLS_AFTER_BIND, AUTO_BIND_NO_TLS, \
     STRING_TYPES, SEQUENCE_TYPES, MOCK_SYNC, MOCK_ASYNC, NTLM, EXTERNAL, DIGEST_MD5, GSSAPI, PLAIN
 
 from .results import RESULT_SUCCESS, RESULT_COMPARE_TRUE, RESULT_COMPARE_FALSE
@@ -134,7 +134,7 @@ class Connection(object):
     :param password: the password for simple authentication
     :type password: str
     :param auto_bind: specify if the bind will be performed automatically when defining the Connection object
-    :type auto_bind: int, can be one of AUTO_BIND_NONE, AUTO_BIND_NO_TLS, AUTO_BIND_TLS_BEFORE_BIND, AUTO_BIND_TLS_AFTER_BIND as specified in ldap3
+    :type auto_bind: int, can be one of AUTO_BIND_DEFAULT, AUTO_BIND_NONE, AUTO_BIND_NO_TLS, AUTO_BIND_TLS_BEFORE_BIND, AUTO_BIND_TLS_AFTER_BIND as specified in ldap3
     :param version: LDAP version, default to 3
     :type version: int
     :param authentication: type of authentication
@@ -163,6 +163,8 @@ class Connection(object):
     :type pool_size: int
     :param pool_lifetime: pool lifetime for pooled strategies
     :type pool_lifetime: int
+    :param cred_store: credential store for gssapi
+    :type cred_store: dict
     :param use_referral_cache: keep referral connections open and reuse them
     :type use_referral_cache: bool
     :param auto_escape: automatic escaping of filter values
@@ -174,7 +176,7 @@ class Connection(object):
                  server,
                  user=None,
                  password=None,
-                 auto_bind=AUTO_BIND_NONE,
+                 auto_bind=AUTO_BIND_DEFAULT,
                  version=3,
                  authentication=None,
                  client_strategy=SYNC,
@@ -190,6 +192,7 @@ class Connection(object):
                  pool_name=None,
                  pool_size=None,
                  pool_lifetime=None,
+                 cred_store=None,
                  fast_decoder=True,
                  receive_timeout=None,
                  return_empty_attributes=True,
@@ -233,7 +236,7 @@ class Connection(object):
             self.closed = True
             self.last_error = None
             if auto_bind is False:  # compatibility with older version where auto_bind was a boolean
-                self.auto_bind = AUTO_BIND_NONE
+                self.auto_bind = AUTO_BIND_DEFAULT
             elif auto_bind is True:
                 self.auto_bind = AUTO_BIND_NO_TLS
             else:
@@ -254,6 +257,7 @@ class Connection(object):
             self.lazy = lazy
             self.pool_name = pool_name if pool_name else conf_default_pool_name
             self.pool_size = pool_size
+            self.cred_store = cred_store
             self.pool_lifetime = pool_lifetime
             self.pool_keepalive = pool_keepalive
             self.starting_tls = False
@@ -329,7 +333,7 @@ class Connection(object):
                     log(BASIC, 'instantiated Connection: <%r>', self)
 
     def do_auto_bind(self):
-        if self.auto_bind and self.auto_bind != AUTO_BIND_NONE:
+        if self.auto_bind and self.auto_bind not in [AUTO_BIND_NONE, AUTO_BIND_DEFAULT]:
             if log_enabled(BASIC):
                 log(BASIC, 'performing automatic bind for <%s>', self)
             if self.closed:
@@ -387,6 +391,7 @@ class Connection(object):
         r += '' if self.pool_size is None else ', pool_size={0.pool_size!r}'.format(self)
         r += '' if self.pool_lifetime is None else ', pool_lifetime={0.pool_lifetime!r}'.format(self)
         r += '' if self.pool_keepalive is None else ', pool_keepalive={0.pool_keepalive!r}'.format(self)
+        r += '' if self.cred_store is None else (', cred_store=' + repr(self.cred_store))
         r += '' if self.fast_decoder is None else (', fast_decoder=' + ('True' if self.fast_decoder else 'False'))
         r += '' if self.auto_range is None else (', auto_range=' + ('True' if self.auto_range else 'False'))
         r += '' if self.receive_timeout is None else ', receive_timeout={0.receive_timeout!r}'.format(self)
@@ -425,6 +430,7 @@ class Connection(object):
         r += '' if self.pool_size is None else ', pool_size={0.pool_size!r}'.format(self)
         r += '' if self.pool_lifetime is None else ', pool_lifetime={0.pool_lifetime!r}'.format(self)
         r += '' if self.pool_keepalive is None else ', pool_keepalive={0.pool_keepalive!r}'.format(self)
+        r += '' if self.cred_store is None else (', cred_store=' + repr(self.cred_store))
         r += '' if self.fast_decoder is None else (', fast_decoder=' + 'True' if self.fast_decoder else 'False')
         r += '' if self.auto_range is None else (', auto_range=' + ('True' if self.auto_range else 'False'))
         r += '' if self.receive_timeout is None else ', receive_timeout={0.receive_timeout!r}'.format(self)
@@ -466,10 +472,13 @@ class Connection(object):
     def __enter__(self):
         with self.connection_lock:
             self._context_state.append((self.bound, self.closed))  # save status out of context as a tuple in a list
-            if self.closed:
-                self.open()
-            if not self.bound:
-                self.bind()
+            if self.auto_bind != AUTO_BIND_NONE:
+                if self.auto_bind == AUTO_BIND_DEFAULT:
+                    self.auto_bind = AUTO_BIND_NO_TLS
+                if self.closed:
+                    self.open()
+                if not self.bound:
+                    self.bind()
 
             return self
 
@@ -747,6 +756,9 @@ class Connection(object):
 
                 if controls is None:
                     controls = []
+                else:
+                    # Copy the controls to prevent modifying the original object
+                    controls = list(controls)
                 controls.append(paged_search_control(paged_criticality, paged_size, paged_cookie))
 
             if self.server and self.server.schema and self.check_names:
@@ -769,6 +781,7 @@ class Connection(object):
                                        self.auto_escape if auto_escape is None else auto_escape,
                                        self.auto_encode,
                                        self.server.schema if self.server else None,
+                                       validator=self.server.custom_validator,
                                        check_names=self.check_names)
             if log_enabled(PROTOCOL):
                 log(PROTOCOL, 'SEARCH request <%s> sent via <%s>', search_request_to_dict(request), self)
@@ -867,6 +880,7 @@ class Connection(object):
         if log_enabled(BASIC):
             log(BASIC, 'start ADD operation via <%s>', self)
         self.last_error = None
+        _attributes = deepcopy(attributes)  # dict could change when adding objectClass values
         if self.check_names:
             dn = safe_dn(dn)
             if log_enabled(EXTENDED):
@@ -881,33 +895,33 @@ class Connection(object):
                 parm_object_class = list(object_class) if isinstance(object_class, SEQUENCE_TYPES) else [object_class]
 
             object_class_attr_name = ''
-            if attributes:
-                for attr in attributes:
+            if _attributes:
+                for attr in _attributes:
                     if attr.lower() == 'objectclass':
                         object_class_attr_name = attr
-                        attr_object_class = list(attributes[object_class_attr_name]) if isinstance(attributes[object_class_attr_name], SEQUENCE_TYPES) else [attributes[object_class_attr_name]]
+                        attr_object_class = list(_attributes[object_class_attr_name]) if isinstance(_attributes[object_class_attr_name], SEQUENCE_TYPES) else [_attributes[object_class_attr_name]]
                         break
             else:
-                attributes = dict()
+                _attributes = dict()
 
             if not object_class_attr_name:
                 object_class_attr_name = 'objectClass'
 
             attr_object_class = [to_unicode(object_class) for object_class in attr_object_class]  # converts objectclass to unicode in case of bytes value
-            attributes[object_class_attr_name] = reduce(lambda x, y: x + [y] if y not in x else x, parm_object_class + attr_object_class, [])  # remove duplicate ObjectClasses
+            _attributes[object_class_attr_name] = reduce(lambda x, y: x + [y] if y not in x else x, parm_object_class + attr_object_class, [])  # remove duplicate ObjectClasses
 
-            if not attributes[object_class_attr_name]:
+            if not _attributes[object_class_attr_name]:
                 self.last_error = 'objectClass attribute is mandatory'
                 if log_enabled(ERROR):
                     log(ERROR, '%s for <%s>', self.last_error, self)
                 raise LDAPObjectClassError(self.last_error)
 
             if self.server and self.server.schema and self.check_names:
-                for object_class_name in attributes[object_class_attr_name]:
+                for object_class_name in _attributes[object_class_attr_name]:
                     if object_class_name.lower() not in conf_classes_excluded_from_check and object_class_name not in self.server.schema.object_classes:
                         raise LDAPObjectClassError('invalid object class ' + str(object_class_name))
 
-                for attribute_name in attributes:
+                for attribute_name in _attributes:
                     if ';' in attribute_name:  # remove tags for checking
                         attribute_name_to_check = attribute_name.split(';')[0]
                     else:
@@ -916,7 +930,7 @@ class Connection(object):
                     if attribute_name_to_check.lower() not in conf_attributes_excluded_from_check and attribute_name_to_check not in self.server.schema.attribute_types:
                         raise LDAPAttributeError('invalid attribute type ' + attribute_name_to_check)
 
-            request = add_operation(dn, attributes, self.auto_encode, self.server.schema if self.server else None, validator=self.server.custom_validator if self.server else None, check_names=self.check_names)
+            request = add_operation(dn, _attributes, self.auto_encode, self.server.schema if self.server else None, validator=self.server.custom_validator if self.server else None, check_names=self.check_names)
             if log_enabled(PROTOCOL):
                 log(PROTOCOL, 'ADD request <%s> sent via <%s>', add_request_to_dict(request), self)
             response = self.post_send_single_response(self.send('addRequest', request, controls))
@@ -1212,6 +1226,8 @@ class Connection(object):
                     log(BASIC, 'deferring START TLS for <%s>', self)
             else:
                 self._deferred_start_tls = False
+                if self.closed:
+                    self.open()
                 if self.server.tls.start_tls(self) and self.strategy.sync:  # for asynchronous connections _start_tls is run by the strategy
                     if read_server_info:
                         self.refresh_server_info()  # refresh server info as per RFC4515 (3.1.5)
@@ -1261,54 +1277,58 @@ class Connection(object):
             result = None
             if not self.sasl_in_progress:
                 self.sasl_in_progress = True  # ntlm is same of sasl authentication
-                # additional import for NTLM
-                from ..utils.ntlm import NtlmClient
-                domain_name, user_name = self.user.split('\\', 1)
-                ntlm_client = NtlmClient(user_name=user_name, domain=domain_name, password=self.password)
+                try:
+                    # additional import for NTLM
+                    from ..utils.ntlm import NtlmClient
+                    domain_name, user_name = self.user.split('\\', 1)
+                    ntlm_client = NtlmClient(user_name=user_name, domain=domain_name, password=self.password)
 
-                # as per https://msdn.microsoft.com/en-us/library/cc223501.aspx
-                # send a sicilyPackageDiscovery request (in the bindRequest)
-                request = bind_operation(self.version, 'SICILY_PACKAGE_DISCOVERY', ntlm_client)
-                if log_enabled(PROTOCOL):
-                    log(PROTOCOL, 'NTLM SICILY PACKAGE DISCOVERY request sent via <%s>', self)
-                response = self.post_send_single_response(self.send('bindRequest', request, controls))
-                if not self.strategy.sync:
-                    _, result = self.get_response(response)
-                else:
-                    result = response[0]
-                if 'server_creds' in result:
-                    sicily_packages = result['server_creds'].decode('ascii').split(';')
-                    if 'NTLM' in sicily_packages:  # NTLM available on server
-                        request = bind_operation(self.version, 'SICILY_NEGOTIATE_NTLM', ntlm_client)
-                        if log_enabled(PROTOCOL):
-                            log(PROTOCOL, 'NTLM SICILY NEGOTIATE request sent via <%s>', self)
-                        response = self.post_send_single_response(self.send('bindRequest', request, controls))
-                        if not self.strategy.sync:
-                            _, result = self.get_response(response)
-                        else:
+                    # as per https://msdn.microsoft.com/en-us/library/cc223501.aspx
+                    # send a sicilyPackageDiscovery request (in the bindRequest)
+                    request = bind_operation(self.version, 'SICILY_PACKAGE_DISCOVERY', ntlm_client)
+                    if log_enabled(PROTOCOL):
+                        log(PROTOCOL, 'NTLM SICILY PACKAGE DISCOVERY request sent via <%s>', self)
+                    response = self.post_send_single_response(self.send('bindRequest', request, controls))
+                    if not self.strategy.sync:
+                        _, result = self.get_response(response)
+                    else:
+                        result = response[0]
+                    if 'server_creds' in result:
+                        sicily_packages = result['server_creds'].decode('ascii').split(';')
+                        if 'NTLM' in sicily_packages:  # NTLM available on server
+                            request = bind_operation(self.version, 'SICILY_NEGOTIATE_NTLM', ntlm_client)
                             if log_enabled(PROTOCOL):
-                                log(PROTOCOL, 'NTLM SICILY NEGOTIATE response <%s> received via <%s>', response[0], self)
-                            result = response[0]
-
-                        if result['result'] == RESULT_SUCCESS:
-                            request = bind_operation(self.version, 'SICILY_RESPONSE_NTLM', ntlm_client, result['server_creds'])
-                            if log_enabled(PROTOCOL):
-                                log(PROTOCOL, 'NTLM SICILY RESPONSE NTLM request sent via <%s>', self)
+                                log(PROTOCOL, 'NTLM SICILY NEGOTIATE request sent via <%s>', self)
                             response = self.post_send_single_response(self.send('bindRequest', request, controls))
                             if not self.strategy.sync:
                                 _, result = self.get_response(response)
                             else:
                                 if log_enabled(PROTOCOL):
-                                    log(PROTOCOL, 'NTLM BIND response <%s> received via <%s>', response[0], self)
+                                    log(PROTOCOL, 'NTLM SICILY NEGOTIATE response <%s> received via <%s>', response[0],
+                                        self)
                                 result = response[0]
-                else:
-                    result = None
-                self.sasl_in_progress = False
 
-            if log_enabled(BASIC):
-                log(BASIC, 'done SASL NTLM operation, result <%s>', result)
+                            if result['result'] == RESULT_SUCCESS:
+                                request = bind_operation(self.version, 'SICILY_RESPONSE_NTLM', ntlm_client,
+                                                         result['server_creds'])
+                                if log_enabled(PROTOCOL):
+                                    log(PROTOCOL, 'NTLM SICILY RESPONSE NTLM request sent via <%s>', self)
+                                response = self.post_send_single_response(self.send('bindRequest', request, controls))
+                                if not self.strategy.sync:
+                                    _, result = self.get_response(response)
+                                else:
+                                    if log_enabled(PROTOCOL):
+                                        log(PROTOCOL, 'NTLM BIND response <%s> received via <%s>', response[0], self)
+                                    result = response[0]
+                    else:
+                        result = None
+                finally:
+                    self.sasl_in_progress = False
 
-            return result
+                if log_enabled(BASIC):
+                    log(BASIC, 'done SASL NTLM operation, result <%s>', result)
+
+                return result
 
     def refresh_server_info(self):
         # if self.strategy.no_real_dsa:  # do not refresh for mock strategies
